/*******************************************************************************
 * 
 * jKiwi -- A virtual makeover and hairstyler application
 * Copyright (c) 2007-2008 Dan Mihai Ile, Maria Joao Roque Barbado Leal
 * 
 * This file is part of jKiwi.
 * 
 * jKiwi is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 * 
 * jKiwi is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 * 
 *******************************************************************************/

package dialogs;

import java.awt.AlphaComposite;
import java.awt.BasicStroke;
import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.Component;
import java.awt.Cursor;
import java.awt.Frame;
import java.awt.GradientPaint;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.Paint;
import java.awt.Panel;
import java.awt.RenderingHints;
import java.awt.Stroke;
import java.awt.event.MouseAdapter;
import java.awt.event.MouseEvent;
import java.awt.event.MouseMotionAdapter;
import java.awt.geom.GeneralPath;
import java.awt.image.BufferedImage;
import java.util.ResourceBundle;

import javax.swing.JRootPane;


import org.eclipse.swt.layout.GridLayout;
import org.eclipse.swt.awt.SWT_AWT;
import org.eclipse.swt.events.SelectionAdapter;
import org.eclipse.swt.events.SelectionEvent;
import org.eclipse.swt.graphics.Image;
import org.eclipse.swt.widgets.Display;
import org.eclipse.swt.widgets.Shell;
import org.eclipse.swt.widgets.Composite;
import org.eclipse.swt.SWT;
import org.eclipse.swt.SWTException;
import org.eclipse.swt.widgets.Button;
import org.eclipse.swt.layout.GridData;
import org.eclipse.swt.widgets.Label;
import org.eclipse.swt.graphics.Font;

import core.Loader;
import core.SubImage;

import utils.StockIcon;

/**
 * Instances of this class allow the user to trace
 * a path for the eyes shape using a new dialog
 */
public class EyesTracer {

	/**
	 * Class that does all the SWING/AWT drawing
	 */
	private class DrawingContainer extends Component {

		/**
		 * Every class that extends Component should have a serial version
		 */
		private static final long serialVersionUID = -7492714701283049381L;

		/**
		 * The stroke used for the drawing
		 */
		private final Stroke basicStroke = new BasicStroke(3,1,1);

		/**
		 * The stroke's color
		 */
		private final Color strokeColor = Color.RED;

		/**
		 * Alpha composite used to fill the path with transparent color
		 */
		private final AlphaComposite fillAlphaComposite;

		/**
		 * Gradient that hides the left side of the image 
		 */
		private final GradientPaint leftEyeGradient;

		/**
		 * Gradient that hides the right side of the image 
		 */
		private final GradientPaint rightEyeGradient;

		/**
		 * The geometric path of the left eye
		 */
		private GeneralPath gpLeftEye = null;

		/**
		 * The geometric path of the right eye
		 */
		private GeneralPath gpRightEye = null;

		/**
		 * the background image for the SWING/AWT composite
		 */
		private BufferedImage backgroundImage;

		/**
		 * Constructs a new instance of this class given its background image.
		 * @param backgroundImage the image that will be drawn into the background of the component
		 */
		public DrawingContainer (BufferedImage backgroundImage) {

			//set the background image
			this.backgroundImage = backgroundImage;

			//initialize alpha composite and gradients
			fillAlphaComposite = AlphaComposite.getInstance(AlphaComposite.SRC_OVER,0.3f);
			int centerWidth = backgroundImage.getWidth() / 2;
			int oneQuarter = backgroundImage.getWidth() / 3;
			leftEyeGradient = new GradientPaint(
					centerWidth,				//x coordinate of the first specified Point in user space
					0,							//y coordinate of the first specified Point in user space
					new Color(255,255,255,0),	//Color at the first specified point
					oneQuarter,					//x coordinate of the second specified Point in user space
					0,							//y coordinate of the second specified Point in user space
					new Color(255,255,255,200)	//Color at the second specified point
			);
			rightEyeGradient = new GradientPaint(
					centerWidth,				//x coordinate of the first specified Point in user space
					0,							//y coordinate of the first specified Point in user space
					new Color(255,255,255,0),	//Color at the first specified point
					oneQuarter * 2,				//x coordinate of the second specified Point in user space
					0,							//y coordinate of the second specified Point in user space
					new Color(255,255,255,200)	//Color at the second specified point
			);

			//TODO check for 1st mouse button in all listeners?
			addMouseListener(new MouseAdapter() {
				public void mouseEntered(final MouseEvent evt) {
					//when the mouse enters the drawing area, change the cursor icon
					setCursor (new Cursor (Cursor.HAND_CURSOR));
				}
				public void mouseReleased(final MouseEvent evt) {
					//on mouse release close the path
					if (isFirstStep) {
						gpLeftEye.closePath();
					} else {
						gpRightEye.closePath();
					}

					//a path has been created, enable the next/finish button
					Display dis = compositeAWT_SWT.getDisplay();
					dis.syncExec(new Runnable() {
						public void run() {
							buttonNext.setEnabled(true);
						}
					});
					repaint(); //repaint the drawing area to make changes visible
				}

				public void mousePressed(final MouseEvent evt) {
					//when mouse is pressed create a new path and register the starting point
					if (isFirstStep) {
						gpLeftEye = new GeneralPath();
						gpLeftEye.moveTo(evt.getX(), evt.getY());
					} else {
						gpRightEye = new GeneralPath();
						gpRightEye.moveTo(evt.getX(), evt.getY());
					}
				}
			});

			addMouseMotionListener(new MouseMotionAdapter() {
				public void mouseDragged(final MouseEvent evt) {
					//when the mouse is dragged update the path with the new line
					if (isFirstStep) {
						gpLeftEye.lineTo(evt.getX(), evt.getY());
					} else {
						gpRightEye.lineTo(evt.getX(), evt.getY());
					}
					repaint(); //repaint the drawing area to make changes visible
				}
			});
		}

		/**
		 * all the drawing in the composite is done here
		 */
		public void paint(Graphics g) {
			Graphics2D g2d = (Graphics2D) g;

			//draw the background image
			g2d.drawImage(backgroundImage, 0,0, null);

			//set stroke and drawing quality
			g2d.setRenderingHint(RenderingHints.KEY_ANTIALIASING, RenderingHints.VALUE_ANTIALIAS_ON);
			g2d.setRenderingHint(RenderingHints.KEY_RENDERING, RenderingHints.VALUE_RENDER_QUALITY);
			g2d.setColor(strokeColor);
			g2d.setStroke(basicStroke);

			//draw left eye path
			if (gpLeftEye != null) {
				java.awt.Composite defaultComposite = g2d.getComposite();
				g2d.draw(gpLeftEye);
				g2d.setComposite(fillAlphaComposite);
				g2d.fill(gpLeftEye);
				g2d.setComposite(defaultComposite);
			}

			//draw right eye path
			if (gpRightEye != null) {
				java.awt.Composite defaultComposite = g2d.getComposite();
				g2d.draw(gpRightEye);
				g2d.setComposite(fillAlphaComposite);
				g2d.fill(gpRightEye);
				g2d.setComposite(defaultComposite);
			}

			//draw white gradient
			Paint defaultPaint = g2d.getPaint();
			if (isFirstStep) {
				g2d.setPaint(rightEyeGradient);
				g2d.fillRect(backgroundImage.getWidth()/2, 0, backgroundImage.getWidth(), backgroundImage.getHeight());
			} else {
				g2d.setPaint(leftEyeGradient);
				g2d.fillRect(0, 0, backgroundImage.getWidth()/2, backgroundImage.getHeight());
			}
			g2d.setPaint(defaultPaint);
		}
	}

	//tells where the dialog should look for the two example images
	private final String EYE_LEFT_IMAGE_PATH;
	private final String EYE_RIGHT_IMAGE_PATH;

	/**
	 * The shell used by this dialog
	 */
	private Shell sShell = null;

	/**
	 * The SWING/AWT composite for the drawings
	 */
	private Composite compositeAWT_SWT = null;

	/**
	 * The composite for the help and tips
	 */
	private Composite compositeTips = null;

	/**
	 * The composite for the control buttons
	 */
	private Composite compositeControls = null;


	//The control buttons
	private Button buttonBack = null;
	private Button buttonNext = null;
	private Button buttonCancel = null;

	//separator label
	private Label labelSeparator = null;

	/**
	 * The composite for the example image
	 */
	private Composite compositeExample = null;

	//The two labels used to specify the tips/help
	private Label labelTitle = null;
	private Label labelText = null;

	/**
	 * messages that are to be used by the entire GUI
	 */
	private ResourceBundle messages;

	/**
	 * Flag to indicate if we are in the first or second step
	 */
	private boolean isFirstStep = true;

	/**
	 * The SWING/AWT drawing container
	 */
	private DrawingContainer dc = null;

	/**
	 * The final eye paths
	 */
	private GeneralPath leftEye = null;
	private GeneralPath rightEye = null;

	/**
	 * The eyes image class
	 */
	private final SubImage subImage;

	//TODO check for the exceptions in the comment
	/**
	 * Constructs a new instance of this class given its parent
	 * and the eyes image.
	 *
	 * @param parent a shell which will be the parent of the new instance
	 * @param subImage the image that contains the eyes
	 * @param eyeLeftImagePath the path to the left image file
	 * @param eyeRightImagePath the path to the right image file
	 *
	 * @exception IllegalArgumentException <ul>
	 *    <li>ERROR_NULL_ARGUMENT - if the parent is null</li>
	 * </ul>
	 * @exception SWTException <ul>
	 *    <li>ERROR_THREAD_INVALID_ACCESS - if not called from the thread that created the parent</li>
	 *    <li>ERROR_INVALID_SUBCLASS - if this class is not an allowed subclass</li>
	 * </ul>
	 */
	public EyesTracer(Shell parent, SubImage subImage, String eyeLeftImagePath, String eyeRightImagePath) {

		this.subImage = subImage;
		//initialize the string messages for this dialog
		messages = Loader.getMessages();

		//create the AWT composite with the eyes image
		dc = new DrawingContainer(subImage.getSubImage());

		//set the left and right image paths
		EYE_LEFT_IMAGE_PATH = eyeLeftImagePath;
		EYE_RIGHT_IMAGE_PATH = eyeRightImagePath;

		//create the dialog window
		createSShell(parent);
	}

	/**
	 * This method initializes sShell
	 * @param parent a shell which will be the parent of the new dialog
	 */
	private void createSShell(Shell parent) {

		//grid data for the separator label
		GridData gridDataSeparator = new GridData();
		gridDataSeparator.horizontalAlignment = org.eclipse.swt.layout.GridData.FILL;

		//initialize the shell
		sShell = new Shell(parent, SWT.APPLICATION_MODAL | SWT.DIALOG_TRIM);
		sShell.setText(messages.getString("EYES_TRACER_TITLE"));
		sShell.setLayout(new GridLayout());
		createCompositeAWT_SWT(); //create the AWT composite
		createCompositeTips(); //create the tips composite

		//initialize the separator label
		labelSeparator = new Label(sShell, SWT.HORIZONTAL | SWT.SEPARATOR);
		labelSeparator.setLayoutData(gridDataSeparator);

		createCompositeControls(); //create the lower buttons composite
		sShell.pack();
	}

	/**
	 * This method initializes compositeAWT_SWT
	 */
	private void createCompositeAWT_SWT() {

		//grid data for the composite
		GridData gridDataComposite = new GridData();

		//initialize the composite to hold the SWING/AWT frame
		compositeAWT_SWT = new Composite(sShell, SWT.BORDER | SWT.EMBEDDED);
		compositeAWT_SWT.setLayout(new GridLayout());
		compositeAWT_SWT.setLayoutData(gridDataComposite);

		//create the SWING/AWT frame and containers
		Frame frame = SWT_AWT.new_Frame(compositeAWT_SWT);
		frame.setBackground(Color.white);
		Panel panel = new Panel(new BorderLayout());
		panel.setBackground(Color.white);
		frame.add(panel);
		JRootPane root = new JRootPane();
		panel.add(root);
		java.awt.Container contentPane = root.getContentPane();
		contentPane.setBackground(Color.white);
		contentPane.setLayout(new BorderLayout());
		contentPane.add(dc, BorderLayout.CENTER);

		//resize the composite to the SWING/AWT image size
		gridDataComposite.widthHint = dc.backgroundImage.getWidth() - 2;
		gridDataComposite.heightHint = dc.backgroundImage.getHeight() - 2;
	}

	/**
	 * This method initializes compositeTips	
	 */
	private void createCompositeTips() {

		//grid data for the composite
		GridData gridDataComposite = new GridData();
		gridDataComposite.grabExcessHorizontalSpace = true;
		gridDataComposite.horizontalAlignment = org.eclipse.swt.layout.GridData.FILL;

		//grid data for the label title
		GridData gridDataLabelTitle = new GridData();
		gridDataLabelTitle.horizontalIndent = 5;
		gridDataLabelTitle.verticalAlignment = org.eclipse.swt.layout.GridData.END;
		gridDataLabelTitle.grabExcessVerticalSpace = true;

		//grid data for the label text
		GridData gridDataLabelText = new GridData();
		gridDataLabelText.verticalAlignment = org.eclipse.swt.layout.GridData.BEGINNING;
		gridDataLabelText.grabExcessVerticalSpace = true;
		gridDataLabelText.horizontalIndent = 5;

		//grid layout for the composite
		GridLayout gridLayoutComposite = new GridLayout();
		gridLayoutComposite.numColumns = 2;

		//initialize the tips composite
		compositeTips = new Composite(sShell, SWT.BORDER);
		compositeTips.setLayout(gridLayoutComposite);
		compositeTips.setLayoutData(gridDataComposite);
		createCompositeExample(); //create the image example composite

		//initialize the title and the text labels
		labelTitle = new Label(compositeTips, SWT.NONE);
		labelTitle.setText(messages.getString("EYES_TRACER_FIRST_STEP_TITLE") + ":");
		labelTitle.setLayoutData(gridDataLabelTitle);
		labelTitle.setFont(new Font(Display.getDefault(), "Sans", 14, SWT.BOLD));
		labelText = new Label(compositeTips, SWT.NONE);
		labelText.setText(messages.getString("EYES_TRACER_FIRST_STEP_TEXT"));
		labelText.setLayoutData(gridDataLabelText);
	}

	/**
	 * This method initializes compositeControls
	 */
	private void createCompositeControls() {

		//grid data for the buttons
		GridData gridDataButtonBack = new GridData();
		GridData gridDataButtonNext = new GridData();
		GridData gridDataButtonCancel = new GridData();
		gridDataButtonBack.verticalAlignment = org.eclipse.swt.layout.GridData.FILL;
		gridDataButtonNext.verticalAlignment = org.eclipse.swt.layout.GridData.FILL;
		gridDataButtonCancel.verticalAlignment = org.eclipse.swt.layout.GridData.FILL;

		//grid data and layout for the composite
		GridData gridDataComposite = new GridData();
		gridDataComposite.grabExcessHorizontalSpace = true;
		gridDataComposite.horizontalAlignment = org.eclipse.swt.layout.GridData.END;
		GridLayout gridLayoutComposite = new GridLayout();
		gridLayoutComposite.numColumns = 3;
		gridLayoutComposite.marginHeight = 0;
		gridLayoutComposite.marginWidth = 0;

		//initialize the composite
		compositeControls = new Composite(sShell, SWT.NONE);
		compositeControls.setLayout(gridLayoutComposite);
		compositeControls.setLayoutData(gridDataComposite);

		//initialize the back button
		buttonBack = new Button(compositeControls, SWT.NONE);
		buttonBack.setText(messages.getString("EYES_TRACER_BACK_BUTTON"));
		buttonBack.setImage(new Image(Display.getCurrent(), StockIcon.getIcon(StockIcon.ICON_GO_BACK, StockIcon.SIZE_SMALL_TOOLBAR, true)));
		buttonBack.setLayoutData(gridDataButtonBack);
		buttonBack.setEnabled(false); //by default is not enabled at startup

		//initialize the next/finish button
		buttonNext = new Button(compositeControls, SWT.NONE);
		buttonNext.setText(messages.getString("EYES_TRACER_NEXT_BUTTON"));
		buttonNext.setImage(new Image(Display.getCurrent(), StockIcon.getIcon(StockIcon.ICON_GO_FORWARD, StockIcon.SIZE_SMALL_TOOLBAR, true)));
		buttonNext.setLayoutData(gridDataButtonNext);
		buttonNext.setEnabled(false); //by default is not enabled at startup

		//initialize the cancel button
		buttonCancel = new Button(compositeControls, SWT.NONE);
		buttonCancel.setImage(new Image(Display.getCurrent(), StockIcon.getIcon(StockIcon.ICON_CANCEL, StockIcon.SIZE_LARGE_TOOLBAR, true)));
		buttonCancel.setText(messages.getString("EYES_TRACER_CANCEL_BUTTON"));
		buttonCancel.setLayoutData(gridDataButtonCancel);

		//adjust the width of the buttons to be the same on all of them
		gridDataButtonBack.widthHint = Math.max(Math.max(buttonBack.computeSize(-1, -1).x, buttonNext.computeSize(-1, -1).x), buttonCancel.computeSize(-1, -1).x);
		gridDataButtonNext.widthHint = Math.max(Math.max(buttonBack.computeSize(-1, -1).x, buttonNext.computeSize(-1, -1).x), buttonCancel.computeSize(-1, -1).x);
		gridDataButtonCancel.widthHint = Math.max(Math.max(buttonBack.computeSize(-1, -1).x, buttonNext.computeSize(-1, -1).x), buttonCancel.computeSize(-1, -1).x);

		//buttons listeners
		buttonBack.addSelectionListener(new SelectionAdapter() {
			public void widgetSelected(SelectionEvent e) {
				moveToFirstStep();
			}
		});
		buttonNext.addSelectionListener(new SelectionAdapter() {
			public void widgetSelected(SelectionEvent e) {
				if (isFirstStep) { //next pressed
					moveToSecondStep();
				} else { //finish pressed
					//the paths to be returned as the final result
					leftEye = subImage.getSmallPath(dc.gpLeftEye);
					rightEye = subImage.getSmallPath(dc.gpRightEye);
					sShell.close();
				}
			}
		});
		buttonCancel.addSelectionListener(new SelectionAdapter() {
			public void widgetSelected(SelectionEvent e) {
				sShell.close();
			}
		});
	}

	/**
	 * update the dialog window to the first step
	 */
	private void moveToFirstStep() {
		isFirstStep = true;
		labelTitle.setText(messages.getString("EYES_TRACER_FIRST_STEP_TITLE") + ":");
		labelText.setText(messages.getString("EYES_TRACER_FIRST_STEP_TEXT"));
		labelTitle.pack();
		labelText.pack();
		buttonBack.setEnabled(false);
		buttonNext.setEnabled(true);
		buttonNext.setText(messages.getString("EYES_TRACER_NEXT_BUTTON"));
		buttonNext.setImage(new Image(Display.getCurrent(), StockIcon.getIcon(StockIcon.ICON_GO_FORWARD, StockIcon.SIZE_SMALL_TOOLBAR, true)));
		compositeExample.setBackgroundImage(new Image(Display.getCurrent(), System.getProperty("program.pixmapspath") + EYE_LEFT_IMAGE_PATH));
		dc.repaint(); //also update the AWT frame
	}

	/**
	 * update the dialog window to the second step
	 */
	private void moveToSecondStep() {
		isFirstStep = false;
		labelTitle.setText(messages.getString("EYES_TRACER_SECOND_STEP_TITLE") + ":");
		labelText.setText(messages.getString("EYES_TRACER_SECOND_STEP_TEXT"));
		labelTitle.pack();
		labelText.pack();
		buttonBack.setEnabled(true);
		if (dc.gpRightEye == null) { //set to false only if right eye has no path
			buttonNext.setEnabled(false);
		}
		buttonNext.setText(messages.getString("EYES_TRACER_FINISH_BUTTON"));
		buttonNext.setImage(new Image(Display.getCurrent(), StockIcon.getIcon(StockIcon.ICON_OK, StockIcon.SIZE_LARGE_TOOLBAR, true)));
		compositeExample.setBackgroundImage(new Image(Display.getCurrent(), System.getProperty("program.pixmapspath") + EYE_RIGHT_IMAGE_PATH));
		dc.repaint(); //also update the AWT frame
	}

	/**
	 * This method initializes compositeExample
	 */
	private void createCompositeExample() {

		//grid data for the composite
		GridData gridDataComposite = new GridData();
		gridDataComposite.verticalSpan = 2;
		gridDataComposite.horizontalIndent = 20;

		//initialize the composite
		compositeExample = new Composite(compositeTips, SWT.BORDER);
		compositeExample.setLayoutData(gridDataComposite);
		compositeExample.setLayout(new GridLayout());
		compositeExample.setBackgroundImage(new Image(Display.getCurrent(), System.getProperty("program.pixmapspath") + EYE_LEFT_IMAGE_PATH));

		//resize the composite to the image size
		gridDataComposite.widthHint = compositeExample.getBackgroundImage().getBounds().width - 2;
		gridDataComposite.heightHint = compositeExample.getBackgroundImage().getBounds().height- 2;
	}

	/**
	 * Makes the dialog visible and brings it to the front
	 * of the display.
	 *
	 * @return an array of two GeneralPath, one for the left eye
	 *         and one for the right eye or <code>null</code> if
	 *         the dialog was canceled or an error occurred
	 *
	 * @exception SWTException <ul>
	 *    <li>ERROR_WIDGET_DISPOSED - if the dialog has been disposed</li>
	 *    <li>ERROR_THREAD_INVALID_ACCESS - if not called from the thread that created the dialog</li>
	 * </ul>
	 */
	public GeneralPath[] open() {

		//open the shell and wait to finish
		sShell.open();
		while (!sShell.isDisposed()) {
			if (!sShell.getDisplay().readAndDispatch())
				sShell.getDisplay().sleep();
		}

		//clear paths and image in AWT
		dc.gpLeftEye = null;
		dc.gpRightEye = null;

		//return the final result
		return new GeneralPath[] { leftEye, rightEye };
	}
}